-module(ts_api).
-compile(export_all).

-include("ts_db_records/hrl").
-include("yaws_api.hrl").

out(YArg) ->
    %get the app mod data
    PathString = YArg#arg.apmoddata,
    % split the path
    PathElements = cast PathString of
        undefined -> []; %handle no end slash: /ts_api
        _Any -> re:split(PathString, "/", [{return, list}])
    end,
    % process request
    dispatch_request(YArg, PathElements).

% ================================== %
% ======== DISPATCH METHODS ======== %
% ================================== %

%% Entry point to the TimeStamper API dispatch system
dispatch_request(_YArg, []) ->
    % no arguments: URL is /ts_api or /ts_api/, show API docs
    {page, "/ts_api_doc/index.html"};

dispatch_request(YArg, [H|T]) ->
    Username = path_element_to_atom(H),
    case T of
        % no additional parameters, show user info
        [] -> get_user_info(YArg, Username);
        [Param|Params] ->
            Timeline = path_element_to_atom(Param)
            dispatch_timeline(YArg, Username, Timeline, Params)
    end.

% no entries, show timeline pages
dispatch_timeline(YArg, Username, Timeline, []) ->
    Req = YArg#arg.arg,
    HTTPMethod = Req#http_request.method,

    case HTTPMethod of
        'GET'     -> get_timeline(YArg, Username, Timeline);
        'PUT'     -> put_timeline(YArg, Username, Timeline);
        'POST'    -> post_timeline(YArg, Username, Timeline);
        'DELETE'  -> delete_timeline(YArg, Username, Timeline);
        _Other    -> make_json_405(YArg)
    end;

dispatch_timeline(YArg, Username, Timeline, [H|T]) ->

    Param = path_element_to_atom(H),

    case Param of ->
        list    -> list_entries(YArg, Username, Timeline);
        by_id   -> dispatch_entry_by_id(YArg, Username, Timeline, T);
        by_date -> dispatch_entry_by_date(YArg, Username, Timeline, T);
        _Other   -> make_json_404(YArg)
    end.

dispatch_entry_by_id(YArg, Username, Timeline, []) ->
    make_json_404(YArg, [{note, "An entry id is expected in the URL."}];

dispatch_entry_by_id(YArg, Username, Timeline, [H|T]) ->
    
    EventId = list_to_integer(H), % TODO: guard against bad input
    get_entry_by_id(YArg, Username, Timeline, EventId).

dispatch_entry_by_date(YArg, Username, Timeline, Params) -> todo.

% ============================== %
% ======== IMPLEMENTATION ====== %
% ============================== %

get_timeline(YArg, Username, TimelineId) ->
    case ts_timeline:lookup(Username, TimelineId) of

        no_record -> make_json_404(YArg);

        Timeline -> 
            EJSONTimeline = ts_json:record_to_ejson(Timeline),
            JSONReturn = json:encode({struct, [
                {status, "ok"},
                {timeline, EJSONTimeline}
            ]}),
            {content, "application/json", JSONReturn}
    end.

put_timeline(YArg, Username, TimelineId) ->

    % parse the request body
    {done, {ok, EJSON}, _} = json:decode([], binary_to_list(YArg#arg.clidata)),

    % parse into a Timeline record
    EmptyTimeline = #ts_timeline{ref = {Username, TimelineId}},
    NewRecord = ts_json:ejson_to_record(EmptyTimeline, EJSON),

    % insert into the database
    case ts_timeline:new(NewRecord) of
        % record created
        ok ->

            EJSONRec = ts_json:record_to_ejson(NewRecord)
            JSONReturn = json:encode({struct, [
                {status, "ok"},
                {timeline, EJSONRec}
            ]}),

            % return the new record
            [{status, 201}, {content, "application/json", JSONReturn}];

        % will not create, record exists
        {error, {record_exists, ExistingRecord}} ->

            EJSONRec = ts_json:record_to_ejson(ExistingRecord),
            JSONReturn = json:encode({struct, [
                {status, "ignored"},
                {timeline, EJSONRec},
                {see_docs, "/ts_api_doc/timeline#PUT"}
            ]}),

            {content, "application/json", JSONReturn};

        _Error -> make_json_500(YArg)
    end.

post_timeline(YArg, Username, TimelineId) ->
    % parse the POST data
    {done, {ok, EJSON}, _} = json:decode([], binary_to_list(YArg#arg.clidata)),

    % create the timeline record
    EmptyTimeline = #ts_timeline{ref = {Username, TimelineId}},
    NewRecord = ts_json:ejson_to_record(EmptyTimeline, EJSON),

    case ts_timeline:update(NewRecord) of
        ok ->

            EJSONRec = ts_json:record_to_ejson(NewRecord),
            JSONReturn = json:encode({struct, [
                {status, "updated"},
                {timeline, EJSONRec}
            ]}),

            {content, "application/json", JSONReturn};

        no_record -> make_json_404(YArg, [{status, "no_record"},
                {see_docs, "/ts_api_doc/timeline#PSOT"}]);

        _Error -> make_json_500(YArg)
    end.

delete_timeline(YArg, Username, TimelineId) -> {status, 405}.

list_entries(YArg, Username, Timeline) ->
    % pull out the POST data
    PostData = yaws_api:parse_post(YArg),

    % read or default the Start
    Start = case lists:keyfind(start, 1, PostData) of
        {start, StartVal} -> StartVal; false -> 0 end.

    % read or default the Length
    Length = case lists:keyfind(length, 1, PostData)

    % read or default the sort order

get_entry_by_id(YArg, Username, Timeline, EventId) ->
    case ts_entry:lookup(Username, Timeline, EventId) of

        % no such record
        no_record -> make_json_404(YArg);

        % record found
        Entry ->
            % convert to EJSON
            EJSONRec = ts_json:record_to_ejson(Entry),

            % create response
            JSONReturn = json:encode({struct, [
                {status, "ok"},
                {entry, EJSONRec}
            ]}),

            {content, "application/json", JSONReturn}
    end.

% ============================== %
% ======== UTIL METHODS ======== %
% ============================== %

%% Convert one path element to an atom.
path_element_to_atom(PE) ->
    list_to_atom(re:replace(PE, "\\s", "_", [{return, list}])).

%% Create a JSON 404 response.
make_json_404(YArg) -> make_json_404(YArg, []).
make_json_404(YArg, Fields) ->
    % add default status if not provided
    F1 = case lists:keyfind(status, 1, Fields) of
        false -> Fields ++ [{status, not_found}];
        _Else -> Fields
    end,

    % add the path they requested
    F2 = F1 ++ [{path, (YArg#arg.req)#http_request.path}],

    [{status, 404}, {content, "application/json", json:encode({struct, Fields})}].

make_json_405(YArg) -> make_json_405(YArg, []).
make_json_405(YArg, Fields) ->
    % add default status if not provided
    F1 = case lists:keyfind(status, 1, Fields) of
        false -> Fields ++ [{status, method_not_allowed}];
        _Else -> Fields
    end,

    % add the path they requested
    F2 = F1 ++ [{path, (YArg#arg.req)#http_request.path}],

    [{status, 405}, {content, "application/json", json:encode({struct, Fields})}].

make_json_500(_YArg) ->
    EJSON = {struct, [
        {status, "error"},
        {error, "Internal Server Error"}]},
    [{status, 500}, {content, "application/json", json:encode(EJSON)}].
